# coding=utf-8

from distutils.core import setup

setup(
    name="pvz",
    version="4.0.1",
    author="lmintlcx",
    author_email="lmintlcx@gmail.com",
    url="https://github.com/lmintlcx/pvzscripts",
    license="GPL",
    description="Python vs. Zombies",
    long_description="Plants vs. Zombies TAS Framework",
    platforms=["Windows"],
    python_requires=">=3.4.0",
    packages=["pvz"],
)
